/******************************************************************************
*
* project name:    FIRE
* initial date:    16/06/2000
* authors:         Thomas Nussbaumer (thomas.nussbaumer@gmx.net)
*                  Lionel Debroux (xdanger1@caramail.com)
*
* description:     fire effect demonstration
*
*
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
* NOTE: current implementation uses TIGCC 0.93
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*
* for TI89 compile with:   tigcc -O3 -DUSE_TI89  dissolve.c
* for TI92P compile with:  tigcc -O3 -DUSE_TI92P dissolve.c
*
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.2: 09/02/2002
*       Update for TIGCC 0.93, a few improvements and optimizations by
*       Lionel Debroux / further improvements and a bugfix (!OSdequeue()) by
*       tnussb
*
* v1.1: 28/06/2000
*       Height of backbuffer modified (previously misstyped) and documentation
*       of InitBackBuffer extended to clearify the used mechanism.
*
* v1.0: 16/06/2000
*       Initial version.
*
*******************************************************************************/

// We don't compile both versions at once, they differ very much...
// (It's done in order to decrease size).

#if defined(USE_TI89) && defined(USE_TI92P)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92P)
#define C89_92(x,y) (y)
#else
#error Please specify either -DUSE_TI89 or -DUSE_TI92P on commandline
#endif


#define SAVE_SCREEN         // restore screen after program end
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#include <tigcclib.h>

/*---------------------------------------------------------------------------*/
/* globals                                                                   */
/*---------------------------------------------------------------------------*/
// used as offscreen buffer to generate image
unsigned char* backbuffer[4] = {NULL,NULL,NULL,NULL};


/*===========================================================================*/
/* allocates memory for backup buffer and clears contents                    */
/*                                                                           */
/* NOTE: all 4 backbuffers (2*2planes) are allocated in one piece and        */
/*       splitted up afterwards to save system handles.                      */
/*       if you allocate each plane itself you would to free each plane      */
/*       separately, too. Due to the fact that we just allocate one large    */
/*       piece of memory we have to free just this one at the end of the     */
/*       program                                                             */
/*===========================================================================*/
static inline short InitBackBuffer(void) {
    register short i;

    if (!(backbuffer[0] = (unsigned char*)malloc(4*LCD_SIZE))) return 0;

    for (i=1;i<4;i++) backbuffer[i] = backbuffer[0]+i*LCD_SIZE;
    memset(backbuffer[0],0,4*LCD_SIZE);
    return 1;
}


/*===========================================================================*/
/* copies offscreen image starting from line 40 to visible bitplanes         */
/* (just from line 40 so that the title won't be overwritten!)               */
/*===========================================================================*/
static inline void BlitScreen(void) {
    register short y;

     // C89_92(100,128) => LCD_HEIGHT
    for (y=40;y<C89_92(100,128);y++) {
        memcpy(GetPlane(LIGHT_PLANE)+((y<<5)-(y<<1)),backbuffer[2]+((y<<5)-(y<<1)),LCD_LINE_BYTES);
        memcpy(GetPlane(DARK_PLANE)+((y<<5)-(y<<1)),backbuffer[3]+((y<<5)-(y<<1)),LCD_LINE_BYTES);
    }

    memcpy(backbuffer[0],backbuffer[2],LCD_SIZE);
    memcpy(backbuffer[1],backbuffer[3],LCD_SIZE);
}


/*===========================================================================*/
/* fillup bottom lines in first backbuffer                                   */
/*===========================================================================*/
static inline void FillBottomLines(void) {
    register short x;
    register unsigned char* ptr;

    // C89_92(2910,3750) => (LCD_HEIGHT-3)*30
    memset(backbuffer[0] + C89_92(2910,3750),0xff,90);
    memset(backbuffer[1] + C89_92(2910,3750),0xff,90);

    // C89_92(2880,3720) => (LCD_HEIGHT-4)*30
    memset(backbuffer[0] + C89_92(2880,3720),0xff,30);
    ptr = backbuffer[1] + C89_92(2880,3720);

    for (x=0;x<30;x++) *ptr++ = (unsigned char)random(256);
}


/*===========================================================================*/
/* copies first backbuffer into second backbuffer with one line offset so    */
/* it looks like as scrolling upwards and modify some pixels randomly        */
/* afterwards                                                                */
/*===========================================================================*/
static inline void ScrollUp(void) {
    register unsigned char *p1;
    register unsigned char *p2;
    short                   value;

    // C89_92(2970,3810) => (LCD_HEIGHT-1)*30
    memcpy(backbuffer[2],backbuffer[0]+30,C89_92(2970,3810));
    memcpy(backbuffer[3],backbuffer[1]+30,C89_92(2970,3810));

    // C89_92(2000,2560) => LCD_HEIGHT*2*30/3
    // C89_92(2910,3750) => (LCD_HEIGHT-3)*30
    for (p1=backbuffer[2]+C89_92(2000,2560),
         p2=backbuffer[3]+C89_92(2000,2560);
         p1<backbuffer[2]+C89_92(2910,3750);
         p1++,p2++)
    {
        value = random(700);
        if (value < 255) *p1 &= (unsigned char)value;
        value = random(700);
        if (value < 255) *p2 &= (unsigned char)value;
    }
}


/*===========================================================================*/
/* Hardcore ClearScreen routine, works on both grayscale planes              */
/*===========================================================================*/
void ClearGrayScreen2B(void *light, void *dark);
asm ("xdef ClearGrayScreen2B\n"\
"ClearGrayScreen2B:\n"\
"  movea.l   4(%a7),%a0\n"\
"  movea.l   8(%a7),%a1\n"\
"  movem.l  %d3-%d7/%a2-%a4,-(%a7)\n"\
"  lea      (0xf00,%a0),%a0\n"\
"  lea      (0xf00,%a1),%a1\n"\
"  moveq    #23,%d7\n"\
"  moveq    #0x0,%d0\n"\
"  moveq    #0x0,%d1\n"\
"  moveq    #0x0,%d2\n"\
"  moveq    #0x0,%d3\n"\
"  moveq    #0x0,%d4\n"\
"  moveq    #0x0,%d5\n"\
"  moveq    #0x0,%d6\n"\
"  movea.l  %d0,%a2\n"\
"  movea.l  %d0,%a3\n"\
"  movea.l  %d0,%a4\n"\
"0:\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"\
"  movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"\
"  dbf      %d7, 0b\n"\
"  movem.l  (%a7)+,%d3-%d7/%a2-%a4\n"\
"  rts");


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    char s1[] = "FIRE EFFECT DEMO";
    char s2[] = "(c) TI-Chess Team 02/2002";
    short key;
    register void *kbq=kbd_queue();
    LCD_BUFFER lcd;

    GKeyFlush();

    //---------------------------------------------
    // turn on grayscale graphics subsystem
    //---------------------------------------------
    if (!GrayOn()) return;

    LCD_save(lcd);

    begin:

    //---------------------------------------------
    // try to allocate memory for offscreen buffers
    //---------------------------------------------
    if (!InitBackBuffer()) {
        GrayOff();
        LCD_restore(lcd);
        ST_helpMsg("Not enough memory !");  // will not be seen!!
        return;
    }

    //---------------------------------------------
    // clear both planes
    //---------------------------------------------
    ClearGrayScreen2B(GetPlane(LIGHT_PLANE),GetPlane(DARK_PLANE));

    //---------------------------------------------
    // write title
    //---------------------------------------------
    for (key=0;key<2;key++) {
        // We want the strings to be written in black, so let's draw them into
        // both planes.
        SetPlane(key);
        FontSetSys(F_8x10); // Large font

        // C89_92(15,55) => (LCD_WIDTH - DrawStrWidth(s1,F_8x10))<<1
        DrawStr(C89_92(15,55),10,s1,A_NORMAL);
        FontSetSys(F_6x8);  // Medium font

        // C89_92(4,44) => (LCD_WIDTH - DrawStrWidth(s,F_6x8))<<1
        DrawStr(C89_92(4,44),30,s2,A_NORMAL);
    }

    //---------------------------------------------
    // initialize the random number generator
    //---------------------------------------------
    randomize();

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    OSTimerRestart(APD_TIMER);
    do {
        FillBottomLines();
        ScrollUp();
        BlitScreen();
        if (OSTimerExpired(APD_TIMER)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        if (!OSdequeue(&key,kbq)) {
            // CLEAR will restart the fire effect.
            // Well, a goto in C is horrible, but sometimes useful...
            if (key==KEY_CLEAR) {
                // Free the offscreen buffers.
                // (else we will have memory leaks, 15360 bytes each time).
                free(backbuffer[0]);
                goto begin;
            }

            // Another key causes exit.
            else break;
        }
    }
    while (1);

    //---------------------------------------------
    // cleanup and bye =8-)
    //---------------------------------------------
    GrayOff();
    LCD_restore(lcd);
    free(backbuffer[0]);
    GKeyFlush();
}
